
USE work.p_constants.ALL;

PACKAGE p_dispatch IS                           -- package declaration

  --- Necessary constants and type definitions.

  CONSTANT IW_Max      : INTEGER := 16;         -- number of IW slots
  CONSTANT BR_Max      : INTEGER := 8;          -- Branch Queue size
  CONSTANT R1_Max      : INTEGER := 8;          -- R1 Queue size
  CONSTANT Excp_BR_Max : INTEGER := 2;          -- Exception Branch Queue size


  CONSTANT OPCODE_doit : INTEGER := 16#5c005c00#; -- plain doit opcode


  --- These are EX Unit functional units.
  CONSTANT OP_arith  : INTEGER := 0;
  CONSTANT OP_branch : INTEGER := 1;
  CONSTANT OP_ctrl   : INTEGER := 2;
  CONSTANT OP_logic  : INTEGER := 3;
  CONSTANT OP_mem    : INTEGER := 4;

  --- These are handled inside the IF Unit.
  CONSTANT OP_doit   : INTEGER := 100;
  CONSTANT OP_rte    : INTEGER := 101;
  CONSTANT OP_sync   : INTEGER := 102;
  CONSTANT OP_sync_x : INTEGER := 103;

  --- These are used mostly for decoding.
  CONSTANT OP_trap   : INTEGER := 200;
  CONSTANT OP_ill    : INTEGER := 201;
  CONSTANT OP_mvbr   : INTEGER := 202;


  TYPE iw_slot IS RECORD
    valid          : BOOLEAN;                   -- slot is filled
    tag            : INTEGER;                   -- dispatch tag
    address        : INTEGER;                   -- instruction address
    opcode         : INTEGER;                   -- instruction opcode
    wat            : BOOLEAN;                   -- is a WAT instruction
    single         : BOOLEAN;                   -- is a SINGLE instruction
    issued         : BOOLEAN;                   -- has been issued
    status         : INTEGER;                   -- completion status
    dest           : INTEGER;                   -- dest reg for SB checking
  END RECORD;

  TYPE iw_type IS ARRAY ( 0 TO IW_Max - 1 ) OF iw_slot;


  --- This function returns TRUE if the opcode is a SINGLE type.
  ---
  FUNCTION is_single( opcode : VLBIT_1D(31 DOWNTO 0) ) RETURN BOOLEAN;
    

  --- This removes the specified instruction from the IW.
  ---
  PROCEDURE iw_remove ( VARIABLE iw_index : INOUT INTEGER; -- which to remove
                        VARIABLE iw : INOUT iw_type; -- IW
                        VARIABLE iw_bottom : INOUT INTEGER; -- iw_bottom
                        VARIABLE wat_flag : INOUT BOOLEAN; -- wat flag
                        VARIABLE single_flag : INOUT BOOLEAN ); -- single flag


  --- determines decode information needed to dispatch instructions
  ---
  PROCEDURE decode ( CONSTANT iopcode         : IN  INTEGER;
                     VARIABLE supervisor_mode : IN  BOOLEAN;
                     VARIABLE exceptions_on   : IN  BOOLEAN;
                     VARIABLE r1_count        : IN  INTEGER;
                     VARIABLE br_count        : IN  INTEGER;
                     VARIABLE new_r1_count    : INOUT INTEGER;
                     VARIABLE new_br_count    : INOUT INTEGER;
                     VARIABLE status          : INOUT INTEGER;
                     VARIABLE op_type         : INOUT INTEGER;
                     VARIABLE single          : INOUT BOOLEAN;
                     VARIABLE wat             : INOUT BOOLEAN;
                     VARIABLE vA              : INOUT BOOLEAN;
                     VARIABLE srcA            : INOUT INTEGER;
                     VARIABLE vB              : INOUT BOOLEAN;
                     VARIABLE srcB            : INOUT INTEGER;
                     VARIABLE vD              : INOUT BOOLEAN;
                     VARIABLE srcD            : INOUT INTEGER;
                     VARIABLE dest            : INOUT INTEGER );

  --- packs variable values into RF operand request vector
  ---
  ---    17   operand S requested
  --- 16-12   operand S register number
  ---    11   operand A requested
  ---  10-6   operand A register number
  ---     5   operand B requested
  ---   4-0   operand B register number
  ---
  FUNCTION rf_encode ( CONSTANT vA    : BOOLEAN;
                       CONSTANT srcA  : INTEGER;
                       CONSTANT vB    : BOOLEAN;
                       CONSTANT srcB  : INTEGER;
                       CONSTANT vD    : BOOLEAN;
                       CONSTANT srcD  : INTEGER )
  RETURN VLBIT_1D;


END p_dispatch;                                 -- end of package declaration



PACKAGE BODY p_dispatch IS                      -- package body is here


  --- Define a table to help with decoding. This table is produced from the
  --- opcodes.tex file in the write-up, by running it through a Perl script. It
  --- doesn't need to mention the .d variants, since they are never decoded.
  --- Instead, when instructions are fetched from memory, the .d is used to
  --- insert a DOIT in the IW. If the IW is changed by an exception handler,
  --- any .d instructions must be explicitly expanded into two IW slots by
  --- software.

---- Start of insertion.

  CONSTANT INDEX_doit   : INTEGER := 55;
  CONSTANT INDEX_rte    : INTEGER := 80;
  CONSTANT INDEX_putcr  : INTEGER := 83;

  CONSTANT T            : BOOLEAN := TRUE;
  CONSTANT F            : BOOLEAN := FALSE;

  TYPE decode_table_entry IS RECORD
    bp      : BOOLEAN;                          -- branch producer
    bc      : BOOLEAN;                          -- branch consumer
    rp      : BOOLEAN;                          -- r1 producer
    rcd     : BOOLEAN;                          -- r1 consumer (dest reg)
    rca     : BOOLEAN;                          -- r1 consumer (src reg A)
    rcb     : BOOLEAN;                          -- r1 consumer (src reg B)
    dr      : BOOLEAN;                          -- destination register
    wat     : BOOLEAN;                          -- wait-at-top instruction
    single  : BOOLEAN;                          -- single instruction
    usr     : BOOLEAN;                          -- user instruction only
    op_type : INTEGER;                          -- type of instruction
  END RECORD;

  TYPE decode_table_type IS ARRAY ( 0 TO 95 ) OF decode_table_entry;
  CONSTANT decode_table : decode_table_type := (
    --- Fields are: BP BC RP RCd RCa RCb DR WAT SGL USR TYPE
    (F,F,F,F,T,F,T,F,F,F,OP_logic ),   --   0  and rd,ra,imm16
    (F,F,F,F,T,F,T,F,F,F,OP_logic ),   --   1  and.u rd,ra,imm16
    (F,F,F,F,T,F,T,F,F,F,OP_logic ),   --   2  mask rd,ra,imm16
    (F,F,F,F,T,F,T,F,F,F,OP_logic ),   --   3  mask.u rd,ra,imm16
    (F,F,F,F,T,F,T,F,F,F,OP_logic ),   --   4  or rd,ra,imm16
    (F,F,F,F,T,F,T,F,F,F,OP_logic ),   --   5  or.u rd,ra,imm16
    (F,F,F,F,T,F,T,F,F,F,OP_logic ),   --   6  xor rd,ra,imm16
    (F,F,F,F,T,F,T,F,F,F,OP_logic ),   --   7  xor.u rd,ra,imm16
    (F,F,F,F,T,F,T,F,F,F,OP_arith ),   --   8  add rd,ra,imm16
    (F,F,F,F,T,F,T,F,F,F,OP_arith ),   --   9  addu rd,ra,imm16
    (F,F,F,F,T,F,T,F,F,F,OP_arith ),   --  10  div rd,ra,imm16
    (F,F,F,F,T,F,T,F,F,F,OP_arith ),   --  11  divu rd,ra,imm16
    (F,F,F,F,T,F,T,F,F,F,OP_arith ),   --  12  mul rd,ra,imm16
    (F,F,F,F,T,F,T,F,F,F,OP_arith ),   --  13  cmp rd,ra,imm16
    (F,F,F,F,T,F,T,F,F,F,OP_arith ),   --  14  sub rd,ra,imm16
    (F,F,F,F,T,F,T,F,F,F,OP_arith ),   --  15  subu rd,ra,imm16
    (T,F,F,F,T,F,F,F,F,F,OP_branch),   --  16  bb0 n,ra,imm16
    (T,F,F,F,T,F,F,F,F,F,OP_branch),   --  17  bb1 n,ra,imm16
    (T,F,F,F,T,F,F,F,F,F,OP_branch),   --  18  ble ra,imm16
    (T,F,F,F,F,F,F,F,F,F,OP_branch),   --  19  br imm26
    (F,F,F,F,F,F,T,F,F,F,OP_ctrl  ),   --  20  mvpc rd,imm16
    (F,F,F,F,F,F,F,F,F,F,OP_trap  ),   --  21  trap imm8
    (F,F,T,T,T,F,T,F,F,F,OP_mem   ),   --  22  xmem rs,ra,imm16
    (F,F,F,F,F,F,F,F,F,F,OP_ill   ),   --  23
    (F,F,T,F,T,F,T,F,F,F,OP_mem   ),   --  24  ld.bu rd,ra,imm16
    (F,F,T,F,T,F,T,F,F,F,OP_mem   ),   --  25  ld.b rd,ra,imm16
    (F,F,T,F,T,F,T,F,F,F,OP_mem   ),   --  26  ld.hu rd,ra,imm16
    (F,F,T,F,T,F,T,F,F,F,OP_mem   ),   --  27  ld.h rd,ra,imm16
    (F,F,T,F,T,F,T,F,F,F,OP_mem   ),   --  28  ld rd,ra,imm16
    (F,F,F,T,T,F,F,F,F,F,OP_mem   ),   --  29  st.b rs,ra,imm16
    (F,F,F,T,T,F,F,F,F,F,OP_mem   ),   --  30  st.h rs,ra,imm16
    (F,F,F,T,T,F,F,F,F,F,OP_mem   ),   --  31  st rs,ra,imm16
    (F,F,F,F,T,T,T,F,F,F,OP_logic ),   --  32  and rd,ra,rb
    (F,F,F,F,T,T,T,F,F,F,OP_logic ),   --  33  and.c rd,ra,rb
    (F,F,F,F,F,F,F,F,F,F,OP_ill   ),   --  34
    (F,F,F,F,F,F,F,F,F,F,OP_ill   ),   --  35
    (F,F,F,F,T,T,T,F,F,F,OP_logic ),   --  36  or rd,ra,rb
    (F,F,F,F,T,T,T,F,F,F,OP_logic ),   --  37  or.c rd,ra,rb
    (F,F,F,F,T,T,T,F,F,F,OP_logic ),   --  38  xor rd,ra,rb
    (F,F,F,F,T,T,T,F,F,F,OP_logic ),   --  39  xor.c rd,ra,rb
    (F,F,F,F,T,T,T,F,F,F,OP_arith ),   --  40  add.io rd,ra,rb
    (F,F,F,F,T,T,T,F,F,F,OP_arith ),   --  41  addu.io rd,ra,rb
    (F,F,F,F,T,T,T,F,F,F,OP_arith ),   --  42  div rd,ra,rb
    (F,F,F,F,T,T,T,F,F,F,OP_arith ),   --  43  divu rd,ra,rb
    (F,F,F,F,T,T,T,F,F,F,OP_arith ),   --  44  mul rd,ra,rb
    (F,F,F,F,T,T,T,F,F,F,OP_arith ),   --  45  cmp rd,ra,rb
    (F,F,F,F,T,T,T,F,F,F,OP_arith ),   --  46  sub.io rd,ra,rb
    (F,F,F,F,T,T,T,F,F,F,OP_arith ),   --  47  subu.io rd,ra,rb
    (T,F,F,F,T,T,F,F,F,F,OP_branch),   --  48  bb0 n,ra,rb
    (T,F,F,F,T,T,F,F,F,F,OP_branch),   --  49  bb1 n,ra,rb
    (T,F,F,F,T,T,F,F,F,F,OP_branch),   --  50  ble ra,rb
    (T,F,F,F,F,T,F,F,F,F,OP_branch),   --  51  br rb
    (F,F,F,F,F,F,F,F,F,F,OP_ill   ),   --  52
    (F,F,F,F,F,F,F,F,F,F,OP_ill   ),   --  53
    (F,F,T,T,T,T,T,F,F,T,OP_mem   ),   --  54  xmem.usr rs,ra,rb
    (F,T,F,F,F,F,F,F,T,F,OP_doit  ),   --  55  doit 
    (F,F,T,F,T,T,T,F,F,T,OP_mem   ),   --  56  ld.h.usr rd,ra,rb
    (F,F,F,F,F,F,F,F,F,F,OP_ill   ),   --  57
    (F,F,F,F,F,F,F,F,F,F,OP_ill   ),   --  58
    (F,F,F,F,F,F,F,F,F,F,OP_ill   ),   --  59
    (F,F,F,T,T,T,F,F,F,T,OP_mem   ),   --  60  st.h.usr rs,ra,rb
    (F,F,F,F,F,F,F,F,F,F,OP_ill   ),   --  61
    (F,F,F,F,F,F,F,F,F,F,OP_ill   ),   --  62
    (F,F,F,F,F,F,F,F,F,F,OP_ill   ),   --  63
    (F,F,F,F,T,T,T,F,F,F,OP_logic ),   --  64  clr rd,ra,rb
    (F,F,F,F,T,T,T,F,F,F,OP_logic ),   --  65  set rd,ra,rb
    (F,F,F,F,T,T,T,F,F,F,OP_logic ),   --  66  ext rd,ra,rb
    (F,F,F,F,T,T,T,F,F,F,OP_logic ),   --  67  extu rd,ra,rb
    (F,F,F,F,T,T,T,F,F,F,OP_logic ),   --  68  mak rd,ra,rb
    (F,F,F,F,T,T,T,F,F,F,OP_logic ),   --  69  rot rd,ra,rb
    (F,F,F,F,F,T,T,F,F,F,OP_logic ),   --  70  ff0 rd,rb
    (F,F,F,F,F,T,T,F,F,F,OP_logic ),   --  71  ff1 rd,rb
    (F,F,F,F,T,F,T,F,F,F,OP_logic ),   --  72  clr rd,ra,w5<o5>
    (F,F,F,F,T,F,T,F,F,F,OP_logic ),   --  73  set rd,ra,w5<o5>
    (F,F,F,F,T,F,T,F,F,F,OP_logic ),   --  74  ext rd,ra,w5<o5>
    (F,F,F,F,T,F,T,F,F,F,OP_logic ),   --  75  extu rd,ra,w5<o5>
    (F,F,F,F,T,F,T,F,F,F,OP_logic ),   --  76  mak rd,ra,w5<o5>
    (F,F,F,F,T,F,T,F,F,F,OP_logic ),   --  77  rot rd,ra,<o5>
    (F,F,F,F,F,F,F,F,F,F,OP_ill   ),   --  78
    (F,F,F,F,F,F,F,F,F,F,OP_ill   ),   --  79
    (F,F,F,F,F,F,F,T,T,T,OP_rte   ),   --  80  rte 
    (F,F,F,F,F,F,F,F,F,F,OP_ill   ),   --  81
    (F,F,F,F,F,F,T,F,F,T,OP_ctrl  ),   --  82  getcr rd,cr
    (F,F,F,F,T,F,F,T,T,T,OP_ctrl  ),   --  83  putcr cr,ra
    (F,F,F,F,F,F,F,T,F,F,OP_sync  ),   --  84  sync 
    (F,F,F,F,F,F,F,T,F,T,OP_sync_x),   --  85  sync.x 
    (F,T,F,F,F,F,T,F,F,F,OP_mvbr  ),   --  86  mvbr rd
    (T,F,F,F,T,F,F,F,F,F,OP_branch),   --  87  ldbr ra
    (F,F,F,F,F,F,F,F,F,F,OP_ill   ),   --  88
    (F,F,F,F,F,F,F,F,F,F,OP_ill   ),   --  89
    (F,F,F,F,F,F,F,F,F,F,OP_ill   ),   --  90
    (F,F,F,F,F,F,F,F,F,F,OP_ill   ),   --  91
    (F,F,F,F,F,F,F,F,F,F,OP_ill   ),   --  92
    (F,F,F,F,F,F,F,F,F,F,OP_ill   ),   --  93
    (F,F,F,F,F,F,F,F,F,F,OP_ill   ),   --  94
    (F,F,F,F,F,F,F,F,F,F,OP_ill   ));  --  95

---- End of insertion.



  --- LOCAL: This function returns the index number of the given opcode
  ---
  FUNCTION op2index( opcode: VLBIT_1D(31 DOWNTO 0) ) RETURN INTEGER IS
    VARIABLE major : INTEGER;
    VARIABLE minor : INTEGER;
    VARIABLE index : INTEGER;
  BEGIN
    
    major := v1d2int( opcode( 30 DOWNTO 26 ) ); -- ignore bit 31
    minor := v1d2int( opcode( 15 DOWNTO 10 ) );

    IF major = 23 THEN
      index := 32 + minor;
    ELSE
      index := major;
    END IF;
        
    RETURN index;

  END op2index;




  --- This function returns TRUE if the opcode is a SINGLE type.
  ---
  FUNCTION is_single( opcode : VLBIT_1D(31 DOWNTO 0) ) RETURN BOOLEAN IS
    VARIABLE index : INTEGER;
  BEGIN

    index := op2index( opcode );

    RETURN decode_table(index).single;
    
  END is_single;



  --- This removes the specified instruction from the IW.
  ---
  PROCEDURE iw_remove ( VARIABLE iw_index : INOUT INTEGER; -- which to remove
                        VARIABLE iw : INOUT iw_type; -- IW
                        VARIABLE iw_bottom : INOUT INTEGER; -- iw_bottom
                        VARIABLE wat_flag : INOUT BOOLEAN; -- wat flag
                        VARIABLE single_flag : INOUT BOOLEAN ) --  single flag
  IS 
  BEGIN

    IF iw(iw_index).wat THEN                    -- clear WAT if appropriate
      wat_flag := FALSE;
    END IF;

    IF iw(iw_index).single THEN                 -- clear SINGLE if appropriate
      single_flag := FALSE;
    END IF;
            
    --- remove this instruction from the IW, and shift the others up

    FOR i IN (iw_index + 1) TO (iw_bottom - 1) LOOP
      iw(i-1) := iw(i);                         -- copy 'em upward
    END LOOP;
    iw_bottom := iw_bottom - 1;                 -- change last pointer
    iw(iw_bottom).valid := FALSE;               -- zap old last entry

  END iw_remove;



  --- determines decode information needed to dispatch instructions
  ---
  PROCEDURE decode ( CONSTANT iopcode         : IN  INTEGER;
                     VARIABLE supervisor_mode : IN  BOOLEAN;
                     VARIABLE exceptions_on   : IN  BOOLEAN;
                     VARIABLE r1_count        : IN  INTEGER;
                     VARIABLE br_count        : IN  INTEGER;
                     VARIABLE new_r1_count    : INOUT INTEGER;
                     VARIABLE new_br_count    : INOUT INTEGER;
                     VARIABLE status          : INOUT INTEGER;
                     VARIABLE op_type         : INOUT INTEGER;
                     VARIABLE single          : INOUT BOOLEAN;
                     VARIABLE wat             : INOUT BOOLEAN;
                     VARIABLE vA              : INOUT BOOLEAN;
                     VARIABLE srcA            : INOUT INTEGER;
                     VARIABLE vB              : INOUT BOOLEAN;
                     VARIABLE srcB            : INOUT INTEGER;
                     VARIABLE vD              : INOUT BOOLEAN;
                     VARIABLE srcD            : INOUT INTEGER;
                     VARIABLE dest            : INOUT INTEGER )
  IS
    VARIABLE index        : INTEGER;
    VARIABLE opcode	  : VLBIT_1D( 31 DOWNTO 0 );
    VARIABLE tmpi	  : INTEGER;
  BEGIN

    opcode := int2v1d( iopcode );
    index := op2index( opcode );

    --- set up default values for outputs

    new_r1_count := r1_count;
    new_br_count := br_count;
    status := STATUS_none;
    op_type := decode_table(index).op_type;
    single := decode_table(index).single;
    wat := decode_table(index).wat;

    vA := decode_table(index).rca;
    vB := decode_table(index).rcb;
    vD := decode_table(index).rcd;
    srcA := v1d2int( opcode(20 DOWNTO 16) );    -- operand A
    srcB := v1d2int( opcode(4 DOWNTO 0) );      -- operand B
    srcD := v1d2int( opcode(25 DOWNTO 21) );    -- operand D

    IF decode_table(index).dr THEN 
      dest := srcD;
    ELSE
      dest := 0;
    END IF;



    --- do the easy ones first

    IF decode_table(index).op_type = OP_ill THEN -- unimplemented opcode
      status := STATUS_illegal;
      RETURN;                                   -- so quit
    END IF;


    IF ( opcode(31) = '1' AND			-- inserts extra DOIT
	 ( index = INDEX_doit OR		-- but that's not allowed for
	   index = INDEX_rte OR			-- these instructions!
	   index = INDEX_putcr ) )
    THEN
      status := STATUS_illegal;
      RETURN;                                   -- so quit
    END IF;


    IF (NOT supervisor_mode) AND decode_table(index).usr THEN
      status := STATUS_user;                    -- users can't do this
      RETURN;
    END IF;


    IF decode_table(index).op_type = OP_trap THEN

      tmpi := v1d2int( opcode(7 DOWNTO 0) );    -- find trap number

      IF (NOT supervisor_mode) AND tmpi < 128 THEN
        status := STATUS_user;                  -- users only trap above 127
        RETURN;
      END IF;

      status := STATUS_trap + tmpi;             -- indicate trap
      RETURN;

    END IF;
        
    --- check deadlock problems

    IF decode_table(index).bc THEN              -- branch consumer (0 or 1)
      new_br_count := new_br_count - 1;
      IF new_br_count < 0 THEN                  -- deadlock
        status := STATUS_lock_b0;
        new_br_count := br_count;
        new_r1_count := r1_count;
        RETURN;
      END IF;
    END IF;
        
    IF decode_table(index).bp THEN              -- branch producer (0 or 1)
      new_br_count := new_br_count + 1;
      IF exceptions_on THEN                     -- how deep is the queue?
        tmpi := BR_Max;
      ELSE
        tmpi := Excp_BR_Max;
      END IF;
      IF new_br_count > tmpi THEN               -- deadlock
        status := STATUS_lock_b1;
        new_br_count := br_count;
        new_r1_count := r1_count;
        RETURN;
      END IF;
    END IF;
        

    IF decode_table(index).rp THEN              -- R1 producer (0 or 1)
      new_r1_count := new_r1_count + 1;
      IF new_r1_count > R1_Max THEN             -- deadlock (no R1 excp queue)
        status := STATUS_lock_r1;
        new_br_count := br_count;
        new_r1_count := r1_count;
        RETURN;
      END IF;
    END IF;
   

    --- Must look at three fields to determine R1 consumer counts.

    IF vA AND srcA = 1 THEN
      new_r1_count := new_r1_count - 1;
    END IF;
    IF vB AND srcB = 1 THEN
      new_r1_count := new_r1_count - 1;
    END IF;
    IF vD AND srcD = 1 THEN
      new_r1_count := new_r1_count - 1;
    END IF;

    IF new_r1_count < 0 THEN                    -- deadlock
      status := STATUS_lock_r0;
      new_br_count := br_count;
      new_r1_count := r1_count;
      RETURN;
    END IF;

    --- By getting this far, I've decoded everything with no faults.

  END decode;



  --- packs variable values into RF operand request vector
  ---
  ---    17   operand S requested
  --- 16-12   operand S register number
  ---    11   operand A requested
  ---  10-6   operand A register number
  ---     5   operand B requested
  ---   4-0   operand B register number
  ---
  FUNCTION rf_encode ( CONSTANT vA    : BOOLEAN;
                       CONSTANT srcA  : INTEGER;
                       CONSTANT vB    : BOOLEAN;
                       CONSTANT srcB  : INTEGER;
                       CONSTANT vD    : BOOLEAN;
                       CONSTANT srcD  : INTEGER )
  RETURN VLBIT_1D IS
    VARIABLE result : VLBIT_1D( 17 DOWNTO 0 );
    VARIABLE tmpv : VLBIT_1D( 31 DOWNTO 0 );
  BEGIN

    result( 17 ) := boo2vlb( vD );
    tmpv := int2v1d( srcD );
    result( 16 DOWNTO 12 ) := tmpv( 4 DOWNTO 0 );

    result( 11 ) := boo2vlb( vA );
    tmpv := int2v1d( srcA );
    result( 10 DOWNTO 6 ) := tmpv( 4 DOWNTO 0 );

    result( 5 ) := boo2vlb( vB );
    tmpv := int2v1d( srcB );
    result( 4 DOWNTO 0 ) := tmpv( 4 DOWNTO 0 );
    
    RETURN result;

  END rf_encode;

END p_dispatch;


--  *-* Text Annotations Start *-*
-- (-7864 -7842 (markup t face italic))
-- (-14659 -14635 (markup t face italic))
-- (-19028 -19020 (markup t face italic))
-- (-19112 -19111 (markup t face italic))
-- (-19177 -19176 (markup t face italic))
-- (-19246 -19245 (markup t face italic))
-- (-19316 -19314 (markup t face italic))
--  *-* Text Annotations End *-*
